from numbers import Number
from typing import Any, Dict, Sequence, Union

import torch
import torchvision.utils as tv_utils
from torch import nn

import meddlr.ops.complex as cplx
from meddlr.config import CfgNode
from meddlr.config.config import configurable
from meddlr.forward.mri import SenseModel
from meddlr.utils.events import get_event_storage
from meddlr.utils.general import move_to_device

from ..layers.layers2D import ResNet
from .build import META_ARCH_REGISTRY, build_model

__all__ = ["GeneralizedUnrolledCNN"]


@META_ARCH_REGISTRY.register()
class GeneralizedUnrolledCNN(nn.Module):
    """Unrolled compressed sensing model.

    This implementation is adapted from:
    https://github.com/MRSRL/dl-cs

    Reference:
        CM Sandino, JY Cheng, et al. "Compressed Sensing: From Research to
        Clinical Practice with Deep Neural Networks" IEEE Signal Processing
        Magazine, 2020.
    """

    @configurable
    def __init__(
        self,
        blocks: Union[nn.Module, Sequence[nn.Module]],
        step_sizes: Union[float, Sequence[float]] = -2.0,
        fix_step_size: bool = False,
        num_emaps: int = 1,
        vis_period: int = -1,
        num_grad_steps: int = None,
    ):
        """
        Args:
            blocks: A sequence of blocks
            step_sizes: Step size for data consistency prior to each block.
                If a single float is given, the same step size is used for all blocks.
            fix_step_size: Whether to fix the step size to a given value --
                i.e. set to ``True`` to make the step size non-trainable.
            num_emaps: Number of sensitivity maps used to estimate the image.
            vis_period: Number of steps between logging visualizations.
            num_grad_steps: Number of unrolled steps in the network.
                This is deprecated - the number of steps will be determined
                from the length of ``blocks``.
        """
        super().__init__()

        self.resnets = blocks
        if num_grad_steps is None:
            if isinstance(blocks, Sequence) and not isinstance(blocks, nn.ModuleList):
                blocks = nn.ModuleList(blocks)
            if not isinstance(blocks, nn.ModuleList):
                raise TypeError("`blocks` must be a sequence of nn.Modules or a nn.ModuleList")
            num_grad_steps = len(blocks)
            num_repeat_steps = 0
        else:
            if not isinstance(num_grad_steps, int) or num_grad_steps <= 0:
                raise ValueError("`num_grad_steps` must be positive integer")
            num_repeat_steps = num_grad_steps

        if isinstance(step_sizes, Number):
            step_sizes = [
                torch.tensor([step_sizes], dtype=torch.float32) for _ in range(num_grad_steps)
            ]
        else:
            if len(step_sizes) != num_grad_steps:
                raise ValueError(
                    "`step_sizes` must be a single value or a list of the "
                    "same length as `blocks` or `num_grad_steps`"
                )
            step_sizes = [torch.tensor(s) for s in step_sizes]
        if not fix_step_size:
            step_sizes = nn.ParameterList([nn.Parameter(s) for s in step_sizes])
        self.step_sizes: Sequence[Union[torch.Tensor, nn.Parameter]] = step_sizes

        self.num_repeat_steps = num_repeat_steps
        self.num_emaps = num_emaps
        self.vis_period = vis_period

    def visualize_training(
        self, kspace: torch.Tensor, zfs: torch.Tensor, targets: torch.Tensor, preds: torch.Tensor
    ):
        """Visualize kspace data and reconstructions.

        Dimension ``(,2)`` indicates optional dimension for real-valued view of complex tensors.
        For example, a real-valued tensor of shape BxHxWx2 will be interpreted as
        a complex-valued tensor of shape BxHxW.

        Args:
            kspace: The complex-valued kspace. Shape: [batch, height, width, #coils, (,2)].
            zfs: The complex-valued zero-filled images.
                Shape: [batch, height, width, (,2)].
            targets: The complex-valued target (reference) images.
                Shape: [batch, height, width, (,2)].
            preds: The complex-valued predicted images.
                Shape: [batch, height, width, (,2)].
        """
        storage = get_event_storage()

        with torch.no_grad():
            if cplx.is_complex(kspace):
                kspace = torch.view_as_real(kspace)
            kspace = kspace[0, ..., 0, :].unsqueeze(0).cpu()  # calc mask for first coil only
            targets = targets[0, ...].unsqueeze(0).cpu()
            preds = preds[0, ...].unsqueeze(0).cpu()
            zfs = zfs[0, ...].unsqueeze(0).cpu()

            all_images = torch.cat([zfs, preds, targets], dim=2)

            imgs_to_write = {
                "phases": cplx.angle(all_images),
                "images": cplx.abs(all_images),
                "errors": cplx.abs(preds - targets),
                "masks": cplx.get_mask(kspace),
            }

            for name, data in imgs_to_write.items():
                data = data.squeeze(-1).unsqueeze(1)
                data = tv_utils.make_grid(data, nrow=1, padding=1, normalize=True, scale_each=True)
                storage.put_image("train/{}".format(name), data.numpy(), data_format="CHW")

    def forward(self, inputs: Dict[str, Any], return_pp: bool = False, vis_training: bool = False):
        """Reconstructs the image from the kspace.

        Dimension ``(,2)`` indicates optional dimension for real-valued view of complex tensors.
        For example, a real-valued tensor of shape BxHxWx2 will be interpreted as
        a complex-valued tensor of shape BxHxW.

        ``#maps`` refers to the number of sensitivity maps used to estimate the image
        (i.e. ``self.num_emaps``).

        Args:
            inputs: Standard meddlr module input dictionary
                * "kspace": The kspace (typically undersampled).
                  Shape: [batch, height, width, #coils, (,2)].
                * "maps": The sensitivity maps used for SENSE coil combination.
                  Shape: [batch, height, width, #coils, #maps, (,2)].
                * "target" (optional): Target (reference) image.
                  Shape: [batch, height, width, #maps, (,2)].
                * "signal_model" (optional): The signal model. If provided,
                    "maps" will not be used to estimate the signal model.
                    Use with caution.
            return_pp (bool, optional): If `True`, return post-processing
                parameters "mean", "std", and "norm" if included in the input.
            vis_training (bool, optional): If `True`, force visualize training
                on this pass. Can only be `True` if model is in training mode.

        Returns:
            Dict: A standard meddlr output dict
                * "pred": The reconstructed image
                * "target" (optional): The target image.
                    Added if provided in the input.
                * "mean"/"std"/"norm" (optional): Pre-processing parameters.
                    Added if provided in the input.
                * "zf_image": The zero-filled image.
                    Added when model is in eval mode.
        """
        if self.num_repeat_steps > 0:
            conv_blocks = [self.resnets] * self.num_repeat_steps
        else:
            conv_blocks = self.resnets

        if vis_training and not self.training:
            raise ValueError("vis_training is only applicable in training mode.")
        # Need to fetch device at runtime for proper data transfer.
        device = next(conv_blocks[0].parameters()).device
        inputs = move_to_device(inputs, device)
        kspace = inputs["kspace"]
        target = inputs.get("target", None)
        mask = inputs.get("mask", None)
        A = inputs.get("signal_model", None)
        maps = inputs["maps"]
        num_maps_dim = -2 if cplx.is_complex_as_real(maps) else -1
        if self.num_emaps != maps.size()[num_maps_dim] and maps.size()[num_maps_dim] != 1:
            raise ValueError("Incorrect number of ESPIRiT maps! Re-prep data...")

        # Move step sizes to the right device.
        step_sizes = [x.to(device) for x in self.step_sizes]

        if mask is None:
            mask = cplx.get_mask(kspace)
        kspace *= mask

        # Get data dimensions
        dims = tuple(kspace.size())

        # Declare signal model.
        if A is None:
            A = SenseModel(maps, weights=mask)

        # Compute zero-filled image reconstruction
        zf_image = A(kspace, adjoint=True)

        # Begin unrolled proximal gradient descent
        image = zf_image
        for resnet, step_size in zip(conv_blocks, step_sizes):
            # dc update
            grad_x = A(A(image), adjoint=True) - zf_image
            image = image + step_size * grad_x

            # If the image is a complex tensor, we view it as a real image
            # where last dimension has 2 channels (real, imaginary).
            # This may take more time, but is done for backwards compatibility
            # reasons.
            # TODO (arjundd): Fix to auto-detect which version of the model is
            # being used.
            use_cplx = cplx.is_complex(image)
            if use_cplx:
                image = torch.view_as_real(image)

            # prox update
            image = image.reshape(dims[0:3] + (self.num_emaps * 2,)).permute(0, 3, 1, 2)
            if hasattr(resnet, "base_forward") and callable(resnet.base_forward):
                image = resnet.base_forward(image)
            else:
                image = resnet(image)

            image = image.permute(0, 2, 3, 1).reshape(dims[0:3] + (self.num_emaps, 2))
            if not image.is_contiguous():
                image = image.contiguous()
            if use_cplx:
                image = torch.view_as_complex(image)

        output_dict = {"pred": image, "target": target}  # N x Y x Z x 1 x 2  # N x Y x Z x 1 x 2

        if return_pp:
            output_dict.update({k: inputs[k] for k in ["mean", "std", "norm"]})

        if self.training and (vis_training or self.vis_period > 0):
            storage = get_event_storage()
            if vis_training or storage.iter % self.vis_period == 0:
                self.visualize_training(kspace, zf_image, target, image)

        output_dict["zf_image"] = zf_image

        return output_dict

    @classmethod
    def from_config(cls, cfg: CfgNode, **kwargs) -> "GeneralizedUnrolledCNN":
        """Build :cls:`GeneralizedUnrolledCNN` from a config.

        Args:
            cfg: The config.
            kwargs: Keyword arguments to override config-specified parameters.

        Returns:
            Dict[str, Any]: The parameters to pass to the constructor.
        """
        # Extract network parameters
        num_grad_steps = cfg.MODEL.UNROLLED.NUM_UNROLLED_STEPS
        share_weights = cfg.MODEL.UNROLLED.SHARE_WEIGHTS

        # Data dimensions
        num_emaps = cfg.MODEL.UNROLLED.NUM_EMAPS

        # Determine block to use for each unrolled step.
        if cfg.MODEL.UNROLLED.BLOCK_ARCHITECTURE == "ResNet":
            builder = lambda: _build_resblock(cfg)  # noqa: E731
        else:
            # TODO: Fix any inconsistencies between config's IN_CHANNELS
            # and the number of channels that the unrolled net expects.
            mcfg = cfg.clone().defrost()
            mcfg.MODEL.META_ARCHITECTURE = cfg.MODEL.UNROLLED.BLOCK_ARCHITECTURE
            mcfg = mcfg.freeze()
            builder = lambda: build_model(mcfg)  # noqa: E731

        # Declare ResNets and RNNs for each unrolled iteration
        if share_weights:
            blocks = builder()
        else:
            blocks = nn.ModuleList([builder() for _ in range(num_grad_steps)])

        # Step sizes
        step_sizes = cfg.MODEL.UNROLLED.STEP_SIZES
        if len(step_sizes) == 1:
            step_sizes = step_sizes[0]

        out = {
            "blocks": blocks,
            "step_sizes": step_sizes,
            "fix_step_size": cfg.MODEL.UNROLLED.FIX_STEP_SIZE,
            "num_emaps": num_emaps,
            "vis_period": cfg.VIS_PERIOD,
            "num_grad_steps": num_grad_steps if share_weights else None,
        }
        out.update(kwargs)
        return out


def _build_resblock(cfg: CfgNode) -> ResNet:
    """Build the resblock for unrolled network.

    Args:
        cfg (CfgNode): The network configuration.

    Note:
        This is a temporary method used as a base case for building
        unrolled networks with the default resblocks. In the future,
        this will be handled by :func:`meddlr.modeling.meta_arch.build_model`.
    """
    # Data dimensions
    num_emaps = cfg.MODEL.UNROLLED.NUM_EMAPS

    # ResNet parameters
    kernel_size = cfg.MODEL.UNROLLED.KERNEL_SIZE
    if len(kernel_size) == 1:
        kernel_size = kernel_size[0]
    resnet_params = dict(
        num_resblocks=cfg.MODEL.UNROLLED.NUM_RESBLOCKS,
        in_chans=2 * num_emaps,  # complex -> real/imag
        chans=cfg.MODEL.UNROLLED.NUM_FEATURES,
        kernel_size=kernel_size,
        drop_prob=cfg.MODEL.UNROLLED.DROPOUT,
        circular_pad=cfg.MODEL.UNROLLED.PADDING == "circular",
        act_type=cfg.MODEL.UNROLLED.CONV_BLOCK.ACTIVATION,
        norm_type=cfg.MODEL.UNROLLED.CONV_BLOCK.NORM,
        norm_affine=cfg.MODEL.UNROLLED.CONV_BLOCK.NORM_AFFINE,
        order=cfg.MODEL.UNROLLED.CONV_BLOCK.ORDER,
    )

    return ResNet(**resnet_params)



from numbers import Number
from typing import Callable, Dict, Iterable, List, Sequence, Union

import torch

import meddlr.ops.complex as cplx
from meddlr.config.config import CfgNode
from meddlr.data.transforms.transform import Normalizer
from meddlr.evaluation.testing import flatten_results_dict
from meddlr.forward import SenseModel
from meddlr.transforms.build import (
    build_iter_func,
    build_scheduler,
    build_transforms,
    seed_tfm_gens,
)
from meddlr.transforms.gen import RandomTransformChoice
from meddlr.transforms.mixins import DeviceMixin, GeometricMixin
from meddlr.transforms.tf_scheduler import SchedulableMixin, TFScheduler
from meddlr.transforms.transform import NoOpTransform, Transform, TransformList
from meddlr.transforms.transform_gen import TransformGen
from meddlr.utils import env


class MRIReconAugmentor(DeviceMixin):
    """
    An augmentation pipeline that manages the organization, generation, and application
    of deterministic and random transforms for MRI reconstruction.

    This class supports both augmentations modeled after the physics of the MRI acquisition process
    (e.g. noise, patient motion) and standard geometric image augmentations (e.g. rotation,
    translation, etc.).

    Physics-driven augmentations are treated as *invariant* transformations.
    Image-based augmentations are treated as *equivariant* transformations.

    Deterministic transformations are those which are applied in the same manner to all data.
    These should be of the type :cls:`Transform`.
    Random transformations are those which are applied randomly to each example or
    batch of examples. These should be of the type :cls:`TransformGen`.

    Note:
        Certain transformations require complex-valued data.
        Unexpected behavior may arise when using real-valued data.

    Reference:
        A Desai, B Gunel, B Ozturkler, et al.
        VORTEX: Physics-Driven Data Augmentations Using Consistency Training
        for Robust Accelerated MRI Reconstruction.
        https://arxiv.org/abs/2111.02549.
    """

    def __init__(
        self,
        tfms_or_gens: Union[TransformList, Sequence[Union[Transform, TransformGen]]],
        aug_sensitivity_maps: bool = True,
        seed: int = None,
        device: torch.device = None,
    ) -> None:
        """
        Args:
            aug_sensitivity_maps: Whether to apply equivariant, image-based
                transforms to sensitivity maps. Note, invariant, physics-based
                augmentations are not applied to the maps.
            seed: A random seed for the random number generator.
            device: The device to use for the transforms.
                If ``None``, the device will be determined automatically
                based on the data passed in.
        """
        if isinstance(tfms_or_gens, TransformList):
            tfms_or_gens = tfms_or_gens.transforms
        self.tfms_or_gens = tfms_or_gens
        self.aug_sensitivity_maps = aug_sensitivity_maps

        if device is not None:
            self.to(device)
        if seed is None:
            seed_tfm_gens(self.tfms_or_gens, seed=seed)

    def __call__(
        self,
        kspace: torch.Tensor,
        maps: torch.Tensor = None,
        target: torch.Tensor = None,
        normalizer: Normalizer = None,
        mask: Union[bool, torch.Tensor] = None,
        mask_gen: Callable = None,
        skip_tfm: bool = False,
    ):
        """Apply augmentations to the set of recon data.

        Args:
            kspace: A complex-valued tensor of shape ``[batch, height, width, #coils]``.
            maps: A complex-valued tensor of shape ``[batch, height, width, #coils, #maps]``.
                To use single-coil data (i.e. ``#coils == 1``), maps should be a Tensor of all ones.
            target: A complex-valued tensor of shape ``[batch, height, width, #maps]``.
            normalizer: A normalizer to normalize the kspace data after equivariant augmentations.
            mask: Whether to use a mask for the SENSE coil combined image.
                If ``True``, a mask will be determined from the non-zero entries of the kspace data.
                If a tensor, it should be a binary tensor broadcastable to the shape of the kspace
                data.
            mask_gen: A callable that returns undersampled kspace and the undersampling mask.
                This undersampling will occur after image-based, equivariant transformations.
            skip_tfm: Whether to skip applying the transformations.

        Returns:
            Tuple[Dict[str, torch.Tensor], List[Transform], List[Transform]]: A tuple of
                a dictionary of transformed data, a list of deterministic equivariant
                transformations, and a list of deterministic invariant transformations.
        """
        if skip_tfm:
            tfms_equivariant, tfms_invariant = [], []
        else:
            # For now, we assume that transforms generated by RandomTransformChoice
            # return random transforms of the same type (equivariant or invariant).
            # We don't have to filter these transforms as a result.
            transform_gens = [
                x.get_transform() if isinstance(x, RandomTransformChoice) else x
                for x in self.tfms_or_gens
            ]
            if any(isinstance(x, (list, tuple)) for x in transform_gens):
                _temp = []
                for t in transform_gens:
                    if isinstance(t, (list, tuple)):
                        _temp.extend(t)
                    else:
                        _temp.append(t)
                transform_gens = _temp

            tfms_equivariant, tfms_invariant = self._classify_transforms(transform_gens)

        use_img = normalizer is not None or len(tfms_equivariant) > 0

        # SENSE coil combination of the image.
        # Note, RSS reconstruction is not currently supported.
        if use_img:
            if mask is True:
                mask = cplx.get_mask(kspace)
            A = SenseModel(maps, weights=mask)
            img = A(kspace, adjoint=True)

        # Apply equivariant transforms to the image.
        if len(tfms_equivariant) > 0:
            img, target, maps = self._permute_data(img, target, maps, spatial_last=True)
            img, target, maps, tfms_equivariant = self._apply_te(
                tfms_equivariant, img, target, maps
            )
            img, target, maps = self._permute_data(img, target, maps, spatial_last=False)

        # Get multi-coil kspace from SENSE combined image.
        if len(tfms_equivariant) > 0:
            A = SenseModel(maps)
            kspace = A(img)

        if mask_gen is not None:
            kspace, mask = mask_gen(kspace)
            img = SenseModel(maps, weights=mask)(kspace, adjoint=True)

        if normalizer:
            normalized = normalizer.normalize(
                **{"masked_kspace": kspace, "image": img, "target": target, "mask": mask}
            )
            kspace = normalized["masked_kspace"]
            target = normalized["target"]
            mean = normalized["mean"]
            std = normalized["std"]
        else:
            mean, std = None, None

        # Apply invariant transforms.
        # Invariant transforms only impact the input (i.e. kspace).
        if len(tfms_invariant) > 0:
            kspace = self._permute_data(kspace, spatial_last=True)
            kspace, tfms_invariant = self._apply_ti(tfms_invariant, kspace)
            kspace = self._permute_data(kspace, spatial_last=False)

        out = {"kspace": kspace, "maps": maps, "target": target, "mean": mean, "std": std}

        for s in self.schedulers():
            s.step(kspace.shape[0])

        return out, tfms_equivariant, tfms_invariant

    def schedulers(self) -> List[TFScheduler]:
        """Returns list of schedulers that are used for all transforms.

        Returns:
            List[TFScheduler]: List of schedulers.
        """
        schedulers = [
            tfm.schedulers() for tfm in self.tfms_or_gens if isinstance(tfm, SchedulableMixin)
        ]
        return [x for y in schedulers for x in y]

    def get_tfm_gen_params(self, scalars_only: bool = True):
        """Get dictionary of scheduler parameters."""
        schedulers: Dict[str, Sequence[TFScheduler]] = {
            type(tfm).__name__: tfm._get_param_values(use_schedulers=True)
            for tfm in self.tfms_or_gens
            if isinstance(tfm, SchedulableMixin)
        }
        params = {}
        for tfm_name, p in schedulers.items():
            p = flatten_results_dict(p, delimiter=".")
            # Filter out values that are not scalars
            p = {f"{tfm_name}/{k}": v for k, v in p.items()}
            if scalars_only:
                p = {k: v for k, v in p.items() if isinstance(v, Number)}
            params.update(p)
        return params

    def _classify_transforms(self, transform_gens):
        tfms_equivariant = []
        tfms_invariant = []
        for tfm in transform_gens:
            if isinstance(tfm, TransformGen):
                tfm_kind = tfm._base_transform
            else:
                tfm_kind = type(tfm)
            assert issubclass(tfm_kind, Transform)

            if issubclass(tfm_kind, GeometricMixin):
                tfms_equivariant.append(tfm)
            else:
                tfms_invariant.append(tfm)
        return tfms_equivariant, tfms_invariant

    def _permute_data(self, *args, spatial_last: bool = False):
        out = []
        if spatial_last:
            for x in args:
                dims = (0,) + tuple(range(3, x.ndim)) + (1, 2)
                out.append(x.permute(dims))
        else:
            for x in args:
                dims = (0, x.ndim - 2, x.ndim - 1) + tuple(range(1, x.ndim - 2))
                out.append(x.permute(dims))
        return out[0] if len(out) == 1 else tuple(out)

    def _apply_te(
        self,
        tfms_equivariant: Iterable[Union[Transform, TransformGen]],
        image: torch.Tensor,
        target: torch.Tensor,
        maps: torch.Tensor,
    ):
        """Apply equivariant transforms.

        These transforms affect both the input and the target.

        Args:
            tfms_equivariant: Equivariant transforms to apply.
            image: The kspace to apply these transformations to.
            target:

        Returns:
            Tuple[torch.Tensor, TransformList]: The transformed kspace and the list
                of deterministic transformations that were applied.
        """
        tfms = []
        for g in tfms_equivariant:
            tfm: Transform = g.get_transform(image) if isinstance(g, TransformGen) else g
            if isinstance(tfm, NoOpTransform):
                continue
            image = tfm.apply_image(image)
            if target is not None:
                target = tfm.apply_image(target)
            if maps is not None and self.aug_sensitivity_maps:
                maps = tfm.apply_maps(maps)
            tfms.append(tfm)
        return image, target, maps, TransformList(tfms, ignore_no_op=True)

    def _apply_ti(
        self, tfms_invariant: Iterable[Union[Transform, TransformGen]], kspace: torch.Tensor
    ):
        """Apply invariant transforms.

        These transforms affect only the input, not the target.

        Args:
            tfms_invariant: Invariant transforms to apply.
            kspace: The kspace to apply these transformations to.

        Returns:
            Tuple[torch.Tensor, TransformList]: The transformed kspace and the list
                of deterministic transformations that were applied.
        """
        tfms = []
        for g in tfms_invariant:
            tfm: Transform = g.get_transform(kspace) if isinstance(g, TransformGen) else g
            if isinstance(tfm, NoOpTransform):
                continue
            kspace = tfm.apply_kspace(kspace)
            tfms.append(tfm)
        return kspace, TransformList(tfms, ignore_no_op=True)

    def reset(self):
        """Reset all transformation generators."""
        for g in self.tfms_or_gens:
            if isinstance(g, TransformGen):
                g.reset()

    def to(self, device: torch.device) -> "MRIReconAugmentor":
        """Moves transformations to a torch device.

        Args:
            device: The device to move to.

        Returns:
            MRIReconAugmentor: self

        Note:
            This method is called in-place. The returned object is self.
        """
        tfms = [tfm for tfm in self.tfms_or_gens if isinstance(tfm, DeviceMixin)]
        for t in tfms:
            t.to(device)
        return self

    @classmethod
    def from_cfg(
        cls, cfg: CfgNode, aug_kind: str, seed: int = None, device: torch.device = None, **kwargs
    ):
        """Build :cls:`MRIReconAugmentor` from a config.

        TODO (arjundd): Decorate __init__ with @configurable.

        Args:
            cfg: A Config node. See ``meddlr/config/defaults.py`` for a complete list of options.
            aug_kind: The scenario in which these augmentations are being used. One of:
                - ``'aug_train'``: Data augmentation as is done in supervised learning.
                - ``'consistency'``: Data augmentations for consistency training (i.e. VORTEX).
            seed: A random seed to initialize this class.
            device: The device to use for the augmentor.

        Returns:
            MRIReconAugmentor: An augmentor.
        """
        mri_tfm_cfg = None
        assert aug_kind in ("aug_train", "consistency")
        if aug_kind == "aug_train":
            mri_tfm_cfg = cfg.AUG_TRAIN.MRI_RECON
        elif aug_kind == "consistency":
            mri_tfm_cfg = cfg.MODEL.CONSISTENCY.AUG.MRI_RECON

        if seed is None and env.is_repro():
            seed = cfg.SEED

        tfms_or_gens = build_transforms(cfg, mri_tfm_cfg.TRANSFORMS, seed=seed, **kwargs)
        scheduler_p = dict(mri_tfm_cfg.SCHEDULER_P)
        ignore_scheduler = scheduler_p.pop("IGNORE", False)
        if len(scheduler_p) and not ignore_scheduler:
            scheduler_p["params"] = ["p"]
            tfms = [tfm for tfm in tfms_or_gens if isinstance(tfm, TransformGen)]
            for tfm in tfms:
                scheduler = build_scheduler(cfg, scheduler_p, tfm)
                tfm.register_schedulers([scheduler])

        if aug_kind in ("aug_train",) and cfg.DATALOADER.NUM_WORKERS > 0:
            func = build_iter_func(cfg.SOLVER.TRAIN_BATCH_SIZE, cfg.DATALOADER.NUM_WORKERS)
            tfms = [tfm for tfm in tfms_or_gens if isinstance(tfm, SchedulableMixin)]
            for tfm in tfms:
                for s in tfm._schedulers:
                    s._iter_fn = func

        return cls(
            tfms_or_gens,
            aug_sensitivity_maps=mri_tfm_cfg.AUG_SENSITIVITY_MAPS,
            seed=seed,
            device=device,
        )



import contextlib
import re
import time
from typing import Dict, Iterable, Sequence, Tuple, Union

import h5py
import numpy as np
import torch

from meddlr.utils import profiler

__all__ = ["HDF5Manager", "structure_patches", "collect_mask"]


class HDF5Manager:
    """Manager for large number of HDF5 files.

    HDF5 files are useful for storing large datasets but have a few limitations
    even in the read-only setting:

        1. File opening and closing can be slow
        2. Reading can be volatile for large datasets on NFS filesystems

    Some functionality includes:

        1. **Caching**: This class can be used to keep files open (defaults to read-only mode).
           This eliminates the need to constantly open and close files.
        2. **Retrying**: Data fetching will be retried in the event of volatile NFS connections.

    Attributes:
        files (Dict[str, h5py.File]): Dictionary of open HDF5 files.
        cache (bool): If `True`, keep files files open.
        max_attempts (int): Maximum number of attempts to read from a file.
            This is useful when the connection to a remote file system
            like a network file storage (NFS) can be volatile.
        wait_time (float): Time to wait between attempts to read from a file.
    """

    def __init__(
        self, files: Sequence[str] = None, cache=True, max_attempts=5, wait_time: float = 1.0
    ):
        """
        Args:
            files (Sequence[str], optional): Filepaths to h5 files the manage.
            cache (bool, optional): If `True`, keep files files open once accessed.
            max_attempts (int, optional): Maximum number of attempts to read from a file.
            wait_time (float, optional): Time to wait between attempts to read from a file.
        """
        self.files: Dict[str, h5py.File] = {}
        self.cache = cache
        self.max_attempts = max(1, max_attempts)
        self.wait_time = wait_time
        if files and cache:
            for fp in files:
                self.open(fp)

    def open(self, filepath, mode="r", **kwargs):
        """Open and cache file.

        Args:
            filepath (str): Path to HDF5 file
        """
        if filepath in self.files:
            return
        self.files[filepath] = h5py.File(filepath, mode, **kwargs)

    def close(self, filepath, drop=True):
        """Close the open file.

        Args:
            filepath (str): Path to HDF5 file
            drop (bool, optional): If `True`, drop filepath from list of files.
                If `False`, `self.files` will keep a reference to the closed file.
        """
        self.files[filepath].close()
        if drop:
            self.files.pop(filepath)

    @profiler.time_profile()
    def get(self, filepath: str, key: str = None, patch: Union[str, Sequence[slice]] = None):
        """Get dataset from h5 file.

        Args:
            filepath (str): Filepath to fetch data from.
            key (str, optional): HDF5 key. If `None`, returns open file.
            patch (str or Sequence[slice], optional): Slice to apply to dataset.
                If `None`, returns dataset without slicing.
        """
        if patch is not None and key is None:
            raise ValueError("`key` must be specified to use `patch`.")

        for idx in range(self.max_attempts):
            try:
                if filepath in self.files:
                    file = self.files[filepath]
                    return self._load_data(file, key=key, sl=patch)
                else:
                    assert key
                    with h5py.File(filepath, "r") as file:
                        data = self._load_data(file, key=key, sl=patch)
                    return data
            except (OSError, KeyError) as e:
                # Handle input/output errors by waiting and retrying.
                # This issue is common for NFS mounted file systems.
                # https://github.com/theislab/scanpy/issues/1351#issuecomment-668009684
                matches_error = any(
                    (re.search(pattern, str(e)) is not None)
                    for pattern in ["\[Errno 5\]", "errno = 5"]
                )
                if (idx < self.max_attempts - 1) and matches_error:
                    is_cached = filepath in self.files
                    if is_cached:
                        self.close(filepath)
                        time.sleep(self.wait_time)
                        self.open(filepath)
                    else:
                        time.sleep(self.wait_time)
                else:
                    raise e

    @profiler.time_profile()
    @contextlib.contextmanager
    def File(self, filepath, mode="r", **kwargs):
        """Yields HDF5 file.

        Similar to ``with h5py.File(...)`` but with supported caching.
        If caching is enabled, the file will be opened (if it is not already
        open) and kept open when the context is exited.

        Example:

        .. code-block:: python

            h5manager = HDF5Manager()
            fpath = "file.h5"
            with h5manager.File(fpath, "r") as h5file:
                arr = h5file["my_array"][()]
        """
        is_cached = filepath in self.files
        file = self.files[filepath] if is_cached else h5py.File(filepath, mode, **kwargs)
        if self.cache and not is_cached:
            self.files[filepath] = file

        try:
            yield file
        finally:
            if not is_cached and not self.cache:
                file.close()

    yield_file = File

    @profiler.time_profile()
    @contextlib.contextmanager
    def temp_open(self, filepath, mode="r", **kwargs) -> "HDF5Manager":
        """Temporarily opens file in this manager.

        If the file is already opened, this open file will be used.

        Yields:
            HDF5Manager: This HDF5Manager (i.e. ``self``).

        Example:

        .. code-block:: python

            h5manager = HDF5Manager()
            fpath = "file.h5"
            with h5manager.temp_open(fpath, "r"):
                arr = h5manager.get(fpath, "my_array", ())
        """
        is_cached = filepath in self.files
        if not is_cached:
            self.open(filepath, mode, **kwargs)

        try:
            yield self
        finally:
            if not is_cached:
                self.close(filepath)

    def _load_data(self, file, key=None, sl=None):
        """Load data from HDF5 file.

        If ``key`` and ``sl`` are not passed, the file is returned.

        If ``key`` is passed, the dataset is returned without loading
        the underlying data into memory. Equivalent to ``file[key]``.

        If ``sl`` is passed, the dataset is returned after slicing.
        Equivalent to ``file[key][sl]``.

        """
        if not key:
            return file

        val = file[key]
        if sl is None:
            return val

        if isinstance(sl, str):
            assert sl in ["all", "load"]
            sl = ()
        elif isinstance(sl, Iterable):
            sl = tuple(sl)
        return val[sl]

    def __del__(self):
        # Copy keys to allow modifying self.files dict.
        for filepath in list(self.files.keys()):
            self.close(filepath)


def structure_patches(
    patches: Sequence[torch.Tensor],
    coords: Union[Sequence[int], Sequence[Tuple[int]]] = None,
    dims=None,
):
    if coords is None:
        coords = list(patches.keys())
        patches = list(patches.values())

    if len(patches) != len(coords):
        raise ValueError("Got unknown length for coordinates")

    if isinstance(coords[0], int):
        coords = [(x,) for x in coords]

    if len(coords) != len(set(coords)):
        raise ValueError(
            "All coordinates must be unique. " "Overlapping patches are not currently supported."
        )

    if dims is not None:
        if isinstance(dims, int):
            dims = (dims,)
        if len(dims) != len(coords[0]):
            raise ValueError(
                f"Specified {len(dims)} dimensions, but have {len(coords[0])}-d coordinates"
            )
        if len(dims) != len(set(dims)):
            raise ValueError(f"Duplicate dimensions specified - {dims}")

    base_shape = patches[0].shape
    if any(x.shape != base_shape for x in patches):
        raise ValueError("All patches must have the same shape")

    stack_shape = []
    coords_arr = np.asarray(coords)
    assert coords_arr.ndim == 2
    stack_shape = tuple(np.max(coords_arr[:, c]).item() + 1 for c in range(coords_arr.shape[1]))

    struct = np.empty(stack_shape, dtype=np.object)
    for coord, patch in zip(coords, patches):
        struct[coord] = patch
    if any(x is None for x in struct.flatten()):
        raise ValueError("Did not get dense labels.")
    struct = struct.tolist()

    out = _recursive_stack(struct)

    if dims is not None:
        # Reorder dimensions of the structured tensor.
        dims = [out.ndim + d if d < 0 else d for d in dims]
        original_dims_new_order = sorted(set(range(out.ndim)) - set(dims))
        order_dict = {v: k for k, v in zip(range(out.ndim), dims + original_dims_new_order)}
        out = out.permute(tuple(order_dict[k] for k in sorted(order_dict.keys())))

    return out


def collect_mask(
    mask: np.ndarray,
    index: Sequence[Union[int, Sequence[int], int]],
    out_channel_first: bool = True,
):
    """Collect masks by index.

    Collated indices will be summed. For example, `index=(1,(3,4))` will return
    `np.stack(mask[...,1], mask[...,3]+mask[...,4])`.

    TODO: Add support for adding background.

    Args:
        mask (ndarray): A (...)xC array.
        index (Sequence[int]): The index/indices to select in mask.
            If sub-indices are collated, they will be summed.
        out_channel_first (bool, optional): Reorders dimensions of output mask to Cx(...)
    """
    if isinstance(index, int):
        index = (index,)

    if not any(isinstance(idx, Sequence) for idx in index):
        mask = mask[..., index]
    else:
        o_seg = []
        for idx in index:
            c_seg = mask[..., idx]
            if isinstance(idx, Sequence):
                c_seg = np.sum(c_seg, axis=-1)
            o_seg.append(c_seg)
        mask = np.stack(o_seg, axis=-1)

    if out_channel_first:
        last_idx = len(mask.shape) - 1
        mask = np.transpose(mask, (last_idx,) + tuple(range(0, last_idx)))

    return mask


def _recursive_stack(tensors) -> torch.Tensor:
    if all(isinstance(x, torch.Tensor) for x in tensors):
        return torch.stack(tensors, dim=0)
    if all(isinstance(x, list) for x in tensors):
        tensors = [_recursive_stack(x) for x in tensors]
        return torch.stack(tensors, dim=0)
    else:
        raise ValueError("Unknown stack")



import os
import warnings
from typing import Any, Callable, TypeVar

TCallable = TypeVar("TCallable", bound=Callable[..., Any])

# List of statistics of deprecated functions.
# This should only be used by the test suite to find any deprecated functions
# that should be removed for this version.
_TRACK_DEPRECATED_FUNCTION_STATS = os.environ.get("MEDDLR_TEST_DEPRECATED", "").lower() == "true"
_DEPRECATED_FUNCTION_STATS = []


def deprecated(
    reason=None, vdeprecated=None, vremove=None, replacement=None
) -> Callable[[TCallable], TCallable]:
    local_vars = locals()

    def fn(func: TCallable) -> TCallable:
        msg = _get_deprecated_msg(func, reason, vdeprecated, vremove, replacement)
        warnings.warn(msg, DeprecationWarning)
        return func

    if _TRACK_DEPRECATED_FUNCTION_STATS:  # pragma: no cover
        _DEPRECATED_FUNCTION_STATS.append(local_vars)

    return fn


def _get_deprecated_msg(wrapped, reason, vdeprecated, vremoved, replacement=None):
    fmt = "{name} is deprecated"
    if vdeprecated:
        fmt += " since v{vdeprecated}"
    if vremoved:
        fmt += " and will be removed in v{vremoved}"
    fmt += "."

    if reason:
        fmt += " ({reason})"
    if replacement:
        fmt += " -- Use meddlr.{replacement} instead."

    return fmt.format(
        name=wrapped.__name__,
        reason=reason or "",
        vdeprecated=vdeprecated or "",
        vremoved=vremoved or "",
        replacement=replacement or "",
    )
